"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs-extra");
const libpath = require("path");
const qs = require("querystring");
const tough = require("tough-cookie");
const ProgressBar = require("progress");
const colors = require("colors");
const puppeteer = require("puppeteer-core");
const utils_1 = require("chrome-launcher/dist/utils");
const chromeFinder = require("chrome-launcher/dist/chrome-finder");
const axios_1 = require("axios");
const axios_cookiejar_support_1 = require("@3846masa/axios-cookiejar-support");
const uploadInfoTemplate_1 = require("./util/uploadInfoTemplate");
const extractTokensFromDOM_1 = require("./util/extractTokensFromDOM");
const Album_1 = require("./Album");
const Photo_1 = require("./Photo");
const USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36';
class GPhotos {
    /**
     * @example
     * ```js
     * const gphotos = new GPhotos({
     *   username: 'username@gmail.com',
     *   password: 'YOUR_PASSWORD',
     *   options: {
     *     progress: true,
     *   }
     * });
     * ```
     */
    constructor(params) {
        const { username, password, options = {} } = params;
        this.username = username;
        this.password = password;
        const defaultOptions = { silence: true, progress: false, jar: new tough.CookieJar() };
        this.options = Object.assign(defaultOptions, options);
        this.axios = axios_cookiejar_support_1.default(axios_1.default.create({
            headers: {
                'User-Agent': USER_AGENT,
            },
            validateStatus: () => true,
            maxRedirects: 0,
            jar: this.options.jar,
            withCredentials: true,
            responseType: 'text',
            transformResponse: [(data) => data],
        }));
    }
    /** @private */
    sendBatchExecute(queries) {
        return __awaiter(this, void 0, void 0, function* () {
            const postArray = [];
            for (const key of Object.keys(queries)) {
                postArray.push([key, JSON.stringify(queries[key]), null, null]);
            }
            const data = yield this.sendQuery('https://photos.google.com/_/PhotosUi/data/batchexecute', [postArray]);
            const results = JSON.parse(data.substr(4)).filter((entry) => entry[0] === 'wrb.fr');
            const error = results.find((entry) => Array.isArray(entry[entry.length - 2]));
            if (error) {
                throw new Error(`Error batchexecute (error: ${error[error.length - 2][0]}, query: ${error[1]})`);
            }
            return results.reduce((obj, entry) => {
                const key = entry[1];
                obj[key] = JSON.parse(entry[2]);
                return obj;
            }, {});
        });
    }
    /** @private */
    sendQuery(url, query) {
        return __awaiter(this, void 0, void 0, function* () {
            const queryRes = yield this.axios.post(url, qs.stringify({
                'f.req': JSON.stringify(query),
                at: this.params.at,
            }));
            if (queryRes.status !== 200) {
                return Promise.reject(new Error(`${queryRes.statusText}`));
            }
            return queryRes.data;
        });
    }
    /**
     * @example
     * ```js
     * gphotos.login()
     *   .then((gphotos) => {
     *     // do something
     *   })
     *   .catch((err) => {
     *     console.error(err.stack);
     *   });
     * ```
     */
    login() {
        return __awaiter(this, void 0, void 0, function* () {
            const params = yield this.fetchParams().catch(() => __awaiter(this, void 0, void 0, function* () {
                yield this.postLogin();
                return this.fetchParams();
            }));
            this.params = params;
            return this;
        });
    }
    /** @private */
    postLogin() {
        return __awaiter(this, void 0, void 0, function* () {
            const chromePath = process.env.PUPPETEER_EXECUTABLE_PATH || (yield chromeFinder[utils_1.getPlatform()]())[0];
            if (!chromePath) {
                throw new Error('Chrome / Chromium is not installed.');
            }
            const browser = yield puppeteer.launch({
                executablePath: chromePath,
                args: ['--no-sandbox', '--disable-setuid-sandbox'],
            });
            try {
                const context = yield browser.createIncognitoBrowserContext();
                const page = yield context.newPage();
                yield page.setUserAgent(USER_AGENT);
                yield Promise.all([
                    page.waitForSelector('input[type="email"]', { visible: true }),
                    page.goto('https://accounts.google.com/ServiceLogin', { waitUntil: 'networkidle2' }),
                ]);
                const $email = (yield page.$('input[type="email"]'));
                yield $email.type(this.username);
                yield Promise.all([page.waitForSelector('input[type="password"]', { visible: true }), $email.press('Enter')]);
                const $password = (yield page.$('input[type="password"]'));
                yield $password.type(this.password);
                yield Promise.all([page.waitForNavigation({ waitUntil: 'networkidle2' }), $password.press('Enter')]);
                const cookies = yield page.cookies();
                for (const cookie of cookies) {
                    yield new Promise((resolve, reject) => {
                        this.options.jar.setCookie(new tough.Cookie({
                            key: cookie.name,
                            value: cookie.value,
                            expires: new Date(cookie.expires * 1000),
                            domain: cookie.domain.replace(/^\./, ''),
                            path: cookie.path,
                        }), page.url(), {
                            http: cookie.httpOnly,
                            secure: cookie.secure,
                            ignoreError: true,
                        }, (err) => (err ? reject(err) : resolve()));
                    });
                }
            }
            finally {
                yield browser.close();
            }
        });
    }
    /** @private */
    fetchParams() {
        return __awaiter(this, void 0, void 0, function* () {
            const gPhotosTopPageRes = yield this.axios.get('https://photos.google.com');
            if (gPhotosTopPageRes.status !== 200) {
                return Promise.reject(new Error("Can't access to Google Photos"));
            }
            const tokens = extractTokensFromDOM_1.default(gPhotosTopPageRes.data);
            if (!tokens.SNlM0e || !tokens.S06Grb) {
                return Promise.reject(new Error("Can't fetch GPhotos params."));
            }
            const params = {
                at: tokens.SNlM0e,
                userId: tokens.S06Grb,
            };
            return params;
        });
    }
    searchAlbum(albumName) {
        return __awaiter(this, void 0, void 0, function* () {
            albumName = `${albumName}`;
            const checkFilter = (info) => {
                return info.title === albumName || info.id === albumName;
            };
            let albumInfo;
            let cursor;
            do {
                const { list, next } = yield this.fetchAlbumList(cursor);
                albumInfo = list.filter(checkFilter).shift();
                cursor = next;
            } while (!albumInfo && cursor);
            if (!albumInfo) {
                throw new Error(`Album "${albumName}" is not found.`);
            }
            return albumInfo;
        });
    }
    fetchAllAlbumList() {
        return __awaiter(this, void 0, void 0, function* () {
            const albumList = [];
            let cursor;
            do {
                const { list, next } = yield this.fetchAlbumList(cursor);
                albumList.push(...list);
                cursor = next;
            } while (cursor);
            return albumList;
        });
    }
    fetchAlbumList(next) {
        return __awaiter(this, void 0, void 0, function* () {
            const { Z5xsfc: results } = yield this.sendBatchExecute({
                Z5xsfc: [next || null, null, null, null, 1],
            });
            if (!results[0]) {
                return { list: [], next: undefined };
            }
            const albumList = results[0].map((al) => {
                const info = al.pop()['72930366'];
                return new Album_1.default({
                    id: al.shift(),
                    title: info[1],
                    period: {
                        from: new Date(info[2][0]),
                        to: new Date(info[2][1]),
                    },
                    items_count: info[3],
                    gphotos: this,
                });
            });
            return { list: albumList, next: results[1] };
        });
    }
    createAlbum(albumName) {
        return __awaiter(this, void 0, void 0, function* () {
            const latestPhoto = yield this.fetchLatestPhoto();
            const { OXvT9d: results } = yield this.sendBatchExecute({
                OXvT9d: [albumName.toString(), null, 1, [[[latestPhoto.id]]]],
            });
            const [[albumId]] = results;
            const album = yield this.searchAlbum(albumId);
            const insertedPhoto = (yield album.fetchPhotoList()).list[0];
            yield insertedPhoto.removeFromAlbum();
            return album;
        });
    }
    fetchPhotoById(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const photo = new Photo_1.default({ id: id, gphotos: this });
            return yield photo.fetchInfo();
        });
    }
    fetchAllPhotoList() {
        return __awaiter(this, void 0, void 0, function* () {
            const photoList = [];
            let cursor;
            do {
                const { list, next } = yield this.fetchPhotoList(cursor);
                photoList.push(...list);
                cursor = next;
            } while (cursor);
            return photoList;
        });
    }
    fetchPhotoList(next) {
        return __awaiter(this, void 0, void 0, function* () {
            const { lcxiM: results } = yield this.sendBatchExecute({
                lcxiM: [next || null, null, null, null, 1],
            });
            if (!results[0]) {
                return { list: [], next: undefined };
            }
            const photoList = results[0].map((info) => {
                const data = Object.assign(Photo_1.default.parseInfo(info), { gphotos: this });
                return new Photo_1.default(data);
            });
            return { list: photoList, next: results[1] };
        });
    }
    fetchLatestPhoto() {
        return __awaiter(this, void 0, void 0, function* () {
            const latestPhotoList = yield this.fetchPhotoList();
            return latestPhotoList.list[0];
        });
    }
    searchOrCreateAlbum(albumName) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.searchAlbum(albumName).catch(() => this.createAlbum(albumName));
        });
    }
    upload(filePath, _fileName) {
        return __awaiter(this, void 0, void 0, function* () {
            const fileName = _fileName || libpath.basename(filePath);
            const fileStat = yield fs.stat(filePath);
            const fileReadStream = fs.createReadStream(filePath);
            return yield this.uploadFromStream(fileReadStream, fileStat.size, fileName);
        });
    }
    uploadFromStream(stream, size, fileName) {
        return __awaiter(this, void 0, void 0, function* () {
            stream.pause();
            const sendInfo = uploadInfoTemplate_1.default();
            for (let field of sendInfo.createSessionRequest.fields) {
                if (field.external) {
                    field.external.filename = fileName || Date.now().toString(10);
                    field.external.size = size;
                }
                else if (field.inlined) {
                    const name = field.inlined.name;
                    if (name !== 'effective_id' && name !== 'owner_name')
                        continue;
                    field.inlined.content = this.params.userId;
                }
            }
            const serverStatusRes = yield this.axios.post('https://photos.google.com/_/upload/photos/resumable?authuser=0', JSON.stringify(sendInfo), {
                headers: {
                    'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
                },
            });
            if (serverStatusRes.status !== 200) {
                return Promise.reject(new Error(`Server Error: ${serverStatusRes.status}`));
            }
            const serverStatus = JSON.parse(serverStatusRes.data);
            if (!('sessionStatus' in serverStatus)) {
                return Promise.reject(new Error('Server Error: sessionStatus is not found.'));
            }
            const sendUrl = serverStatus.sessionStatus.externalFieldTransfers[0].putInfo.url;
            if (this.options.progress) {
                const progressBar = new ProgressBar(`${colors.green('Uploading')} [:bar] :percent :etas`, {
                    total: size,
                });
                stream.on('open', () => process.stderr.write('\n'));
                stream.on('data', (chunk) => {
                    progressBar.tick(chunk.length);
                });
                stream.on('end', () => process.stderr.write('\n'));
            }
            const resultRes = yield this.axios.post(sendUrl, stream, {
                headers: {
                    'Content-Type': 'application/octet-stream',
                    'X-HTTP-Method-Override': 'PUT',
                },
            });
            const result = JSON.parse(resultRes.data);
            if (result.sessionStatus.state !== 'FINALIZED') {
                return Promise.reject(new Error(`Upload Error: ${result.sessionStatus.state}`));
            }
            const uploadInfo = result.sessionStatus.additionalInfo['uploader_service.GoogleRupioAdditionalInfo'].completionInfo
                .customerSpecificInfo;
            const uploadedPhoto = new Photo_1.default({
                id: uploadInfo.photoMediaKey,
                uploadedAt: new Date(),
                createdAt: new Date(uploadInfo.timestamp * 1000),
                type: uploadInfo.kind,
                title: uploadInfo.title,
                description: uploadInfo.description,
                rawUrl: uploadInfo.url,
                uploadInfo: uploadInfo,
                gphotos: this,
            });
            return uploadedPhoto;
        });
    }
}
exports.default = GPhotos;
